"""TODO: module docstring."""

import abc

import numpy as np

from chemex.spindynamics import util

EXP_DETAILS = {"name": {"type": str}}

CONDITIONS = {
    "h_larmor_frq": {"type": float},
    "temperature": {"type": float},
    "p_total": {"default": None, "type": float},
    "l_total": {"default": None, "type": float},
}


class BaseProfile(metaclass=abc.ABCMeta):
    """TODO: class docstring."""

    def __init__(self, name=None, exp_details=None, model=None):
        """TODO: method docstring."""

        if name is None:
            name = ""

        self.profile_name = name
        self.model = util.parse_model(model)
        self.experiment_name = self.check_exp_details(exp_details, EXP_DETAILS)["name"]
        self.conditions = self.check_exp_details(exp_details, CONDITIONS)
        self.val = None
        self.err = None
        self.data = None
        self.mask = None

    def __len__(self):
        return self.val.size

    @abc.abstractmethod
    def print_profile(self):
        """TODO: method docstring."""
        pass

    @abc.abstractmethod
    def calculate_profile(self, params):
        """TODO: method docstring."""
        pass

    def create_default_parameters(self):
        """TODO: method docstring."""
        pass

    @abc.abstractmethod
    def filter_points(self, params=None):
        """TODO: method docstring."""
        pass

    def calculate_residuals(self, params):
        """Calculate the residuals between the experimental and back-calculated
        values."""

        values = self.calculate_profile(params)
        residuals = (self.val - values) / self.err

        return residuals[self.mask]

    @staticmethod
    def check_par(
        parameters=None, name=None, convert=None, default=None, required=True
    ):
        """Check for experimental parameters and convert them to their appropriate
        type."""
        value = parameters.get(name, default)

        if isinstance(value, str):
            value = value.split()
            if len(value) == 1:
                value = value[0]

        if required and value is None:
            exit(
                "Missing experimental parameter detected. Please set: {:s}".format(name)
            )

        if convert is not None and value is not None:
            try:
                if isinstance(value, list):
                    value = [convert(_) for _ in value]
                else:
                    value = convert(value)
            except ValueError:
                exit(
                    "\nExperimental parameter of wrong type detected. Please make"
                    " sure that {:s} is a {:s}".format(name, str(convert))
                )

        return value

    @staticmethod
    def check_exp_details(exp_details=None, expected=None):
        """Check for experimental parameters and convert them to their appropriate
        type."""

        if expected is None:
            expected = {}

        details = {}
        missing = []

        for name, item in expected.items():
            default = item.get("default")
            value = exp_details.get(name, default)
            required = "default" not in item
            if value is None and required:
                missing.append("'" + name + "'")
            elif isinstance(value, str):
                value = value.split()
            if value is None:
                details[name] = None
            else:
                details[name] = np.array(value, dtype=item.get("type")).reshape(-1)

                if details[name].shape == (1,):
                    details[name] = details[name][0]

        if missing:
            exit(
                "\n  - Missing experimental parameter detected. Please set: "
                "{:s}".format(", ".join(missing))
            )

        return details



"""The chemex module provides the entry point for the chemex script."""

import copy
import os
import shutil
import sys

import matplotlib

matplotlib.use("PDF")
import numpy as np

from chemex import datasets, fitting, parameters, cli, util, __version__


LOGO = r"""
* * * * * * * * * * * * * * * * * * * * * * * * *
*      ________                   ______        *
*     / ____/ /_  ___  ____ ___  / ____/  __    *
*    / /   / __ \/ _ \/ __ `__ \/ __/ | |/_/    *
*   / /___/ / / /  __/ / / / / / /____>  <      *
*   \____/_/ /_/\___/_/ /_/ /_/_____/_/|_|      *
*                                               *
*   Analysis of NMR Chemical Exchange data      *
*                                               *
*   Version: {:<34s} *
*                                               *
* * * * * * * * * * * * * * * * * * * * * * * * *
""".format(
    __version__
)


def make_bootstrap_dataset(data):
    """Create a new dataset to run a bootstrap simulation."""
    data_bs = datasets.DataSet()

    for profile in data:
        data_bs.append(profile.make_bs_profile())

    return data_bs


def make_montecarlo_dataset(data, params):
    """Create a new dataset to run a Monte-Carlo simulation."""
    data_mc = datasets.DataSet()

    for profile in data:
        data_mc.append(profile.make_mc_profile(params=params))

    return data_mc


def read_data(args):
    """Read experimental setup and data."""
    util.header1("Reading Experimental Data")

    data = datasets.DataSet()

    if args.experiments:
        print(("{:<45s} {:<25s} {:<25s}".format("File Name", "Experiment", "Profiles")))
        print(("{:<45s} {:<25s} {:<25s}".format("---------", "----------", "--------")))

        for filename in args.experiments:
            data.add_dataset_from_file(
                filename, args.model, args.res_incl, args.res_excl
            )

    if not data.data:
        sys.exit("\nNo data to fit!\n")

    return data


def write_results(result, data, method, output_dir):
    """Write the results of the fit to output files.

    The files below are created and contain the following information:
      - parameters.fit: fitting parameters and their uncertainties
      - contstraints.fit: expression used for constraining parameters
      - *.dat: experimental and fitted data
      - statistics.fit: statistics for the fit

    """
    util.header1("Writing Results")

    print("\nFile(s):")

    if method:
        shutil.copyfile(method, os.path.join(output_dir, "fitting-method.cfg"))

    parameters.write_par(result.params, output_dir=output_dir)
    parameters.write_constraints(result.params, output_dir=output_dir)
    data.write_to(result.params, output_dir=output_dir)
    fitting.write_statistics(result, path=output_dir)


def plot_results(result, data, output_dir):
    """Plot the experimental and fitted data."""
    from chemex.experiments import plotting

    util.header1("Plotting Data")

    print("\nFile(s):")

    output_dir_plot = os.path.join(output_dir, "Plots")
    util.make_dir(output_dir_plot)

    try:
        plotting.plot_data(data, result.params, output_dir=output_dir_plot)
    except KeyboardInterrupt:
        print(" - Plotting cancelled")

    if result.method == "brute":
        labels = [
            parameters.ParameterName.from_full_name(var).name.upper()
            for var in result.var_names
        ]
        outfile = os.path.join(output_dir, "results_brute.pdf")
        plotting.plot_results_brute(result, varlabels=labels, output=outfile)
        print(("  * {}".format(outfile)))


def fit_write_plot(args, params, data, output_dir):
    """Perform the fit, write the output files and plot the results."""
    result = fitting.run_fit(args.method, params, data, args.fitmethod)

    util.make_dir(output_dir)

    write_results(result, data, args.method, output_dir)

    if not args.noplot:
        plot_results(result, data, output_dir)

    return result


def get_info(args):
    # cli.format_experiment_help(args.types, args.experiments)
    print(args.experiments)


def fit(args):
    # Read experimental setup and data
    data = read_data(args)

    # Create and update initial values of fitting/fixed parameters
    util.header1("Reading Default Parameters")
    params = parameters.create_params(data)

    for name in args.parameters:
        parameters.set_params_from_config_file(params, name)

    # Filter datapoints out if necessary (e.g., on-resonance filter CEST)
    for profile in data:
        profile.filter_points(params)
    data.ndata = sum([len(profile.val) for profile in data])

    # Customize the output directory
    output_dir = args.out_dir if args.out_dir else "./Output"
    if args.res_incl:
        if len(args.res_incl) == 1:
            output_dir = os.path.join(output_dir, args.res_incl.pop().upper())

    result = fit_write_plot(args, params, data, output_dir)

    if args.bs or args.mc:
        if args.bs:
            nmb = args.bs
        else:
            nmb = args.mc

        formatter_output_dir = "".join(["{:0", str(int(np.log10(nmb)) + 1), "d}"])

        for index in range(1, nmb + 1):
            if args.bs:
                data_index = make_bootstrap_dataset(data)
            else:
                data_index = make_montecarlo_dataset(data, result.params)

            output_dir_ = os.path.join(output_dir, formatter_output_dir.format(index))

            params_mc = copy.deepcopy(result.params)

            fit_write_plot(args, params_mc, data_index, output_dir_)


def main():
    """Do all the magic."""
    print(LOGO)

    parser = cli.build_parser(fit)
    args = parser.parse_args()

    if args.commands is None:
        parser.print_help()
    else:
        args.func(args)



"""The util module contains a variety of utility functions."""

import configparser
import os
import sys


def make_dir(path=None):
    """Ensure existence of the directory."""
    try:
        os.makedirs(path)
    except OSError:
        if not os.path.isdir(path):
            raise


def read_cfg_file(filename):
    """Read and parse the experiment configuration file with configparser."""
    config = configparser.ConfigParser(inline_comment_prefixes=("#", ";"))
    config.optionxform = str

    if filename:
        try:
            out = config.read(filename)

            if not out:
                exit(
                    "\nERROR: The file '{}' is empty or does not exist!\n".format(
                        filename
                    )
                )

        except configparser.MissingSectionHeaderError:
            exit(
                "\nERROR: You are missing a section heading in {:s}\n".format(filename)
            )

        except configparser.ParsingError:
            exit(
                "\nERROR: Having trouble reading your parameter file, did you"
                " forget '=' signs?\n{:s}".format(sys.exc_info()[1])
            )

    return config


def normalize_path(working_dir, filename):
    """Normalize the path of a filename relative to a specific directory."""
    if not os.path.isabs(filename):
        path = os.path.join(working_dir, filename)
    else:
        path = filename

    return path


def header1(string):
    """Print a formatted heading."""
    print(("\n".join(["", "", string, "=" * len(string), ""])))


def header2(string):
    """Print a formatted subheading."""
    print(("\n".join(["", string, "-" * len(string), ""])))



"""Pure in-phase CPMG.

Analyzes chemical exchange in the presence of high power 1H CW decoupling during
the CPMG block. This keeps the spin system purely in-phase throughout, and is
calculated using the 6x6, single spin matrix:

[ Ix(a), Iy(a), Iz(a), Ix(b), Iy(b), Iz(b) ]

References
----------

Hansen, Vallurupalli and Kay. Journal of Physical Chemistry B (2008) 112:5898-5904

"""

from functools import reduce
from itertools import cycle, islice

import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t2": {"type": float},
    "pw90": {"type": float},
    "time_equil": {"default": 0.0, "type": float},
    "ncyc_max": {"type": int},
}


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.liouv = basis.Liouvillian(
            system="ixyz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.carrier_i = self.exp_details["carrier"]
        self.liouv.w1_i = 2.0 * np.pi / (4.0 * self.exp_details["pw90"])

        self.detect = self.liouv.detect["iz_a"]

        self.t_cps = {}

        for ncyc in self.ncycs[self.ncycs > 0]:
            self.t_cps[ncyc] = (
                self.exp_details["time_t2"] / (4.0 * ncyc)
                - self.exp_details["pw90"] * 0.75
            )

        self.t_pos_2 = +4.0 * self.exp_details["pw90"] / np.pi
        self.t_neg = -2.0 * self.exp_details["pw90"] / np.pi

        self.delays = [self.t_pos_2, self.t_neg, self.exp_details["time_equil"]]
        self.delays.extend(self.t_cps.values())
        self.delays.extend(
            [self.exp_details["pw90"] * _ for _ in range(self.exp_details["ncyc_max"])]
        )

        self.phase_1 = [0, 0, 1, 3, 0, 0, 3, 1, 0, 0, 3, 1, 0, 0, 1, 3]
        self.phase_2 = [1, 3, 2, 2, 3, 1, 2, 2, 3, 1, 2, 2, 1, 3, 2, 2]
        self.n_phase = len(self.phase_1)

        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_i_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """Calculate the intensity in presence of exchange after a CEST block.

        Parameters
        ----------
        pb : float
            Fractional population of state B.
        kex_ab : float
            Exchange rates between states A and B in /s.
        dw_i_ab : float
            Chemical shift difference between states A and B in rad/s.
        r1_i_a : float
            Longitudinal relaxation rate of states A in /s.
        r2_i_a : float
            Transverse relaxation rate of state A in /s.
        dr2_i_ab : float
            Transverse relaxation rate difference between states A and B in /s.
        cs_i_a : float
            Resonance position of state A in ppm.

        Returns
        -------
        out : float
            Intensity after the CEST block

        """

        self.liouv.update(**parvals)

        # Calculation of all the needed propagators
        pulses = self.liouv.pulses_90_180_i()
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))

        p_180s = [pulses["180px"], pulses["180py"], pulses["180mx"], pulses["180my"]]

        # Simulate the CPMG block as function of ncyc
        mag0 = self.liouv.compute_mag_eq(term="iz", **parvals)

        profile = []

        for ncyc in self.ncycs:

            if ncyc == 0:

                delay = self.exp_details["pw90"] * (self.exp_details["ncyc_max"] - 1)

                p_centre = 0.5 * (
                    pulses["180px"] @ delays[self.t_pos_2] @ pulses["180px"]
                    + pulses["180my"] @ delays[self.t_pos_2] @ pulses["180py"]
                )

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ delays[delay]
                    @ pulses["90my"]
                    @ p_centre
                    @ pulses["90py"]
                    @ mag0
                )

            else:

                delay = self.exp_details["pw90"] * (self.exp_details["ncyc_max"] - ncyc)
                p_t_cp = delays[self.t_cps[ncyc]]
                p_echoes = [p_t_cp @ p_180 @ p_t_cp for p_180 in p_180s]

                p_cp1 = reduce(
                    np.matmul,
                    [
                        p_echoes[phase]
                        for phase in islice(cycle(self.phase_1), 2 * ncyc)
                    ],
                )
                p_cp2 = reduce(
                    np.matmul,
                    [
                        p_echoes[phase]
                        for phase in islice(cycle(self.phase_2), 2 * ncyc)
                    ],
                )

                p_cp = 0.5 * (p_cp1 + p_cp2)

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ delays[delay]
                    @ pulses["90my"]
                    @ delays[self.t_neg]
                    @ p_cp
                    @ delays[self.t_neg]
                    @ pulses["90py"]
                    @ mag0
                )

            profile.append(np.float64(self.liouv.collapse(mag)))

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""
        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]



"""The dataset module contains the code for handling the experimental data."""

import configparser
import copy
import importlib
import operator
import os
import os.path
import sys

import numpy as np

from chemex import util
from chemex.experiments import base_profile


class DataSet(object):
    """DataSet class for handling experimental data."""

    def __init__(self, other=None):
        self.data = []
        self.ndata = 0
        self.chisq_ref = 1e32

        if isinstance(other, DataSet):
            self.data = copy.deepcopy(other.data)
            self.ndata = other.ndata

        elif isinstance(other, base_profile.BaseProfile):
            self.data.append(other)
            self.ndata = len(other)

    def __iter__(self):
        for some_data in self.data:
            yield some_data

    def __add__(self, other):
        data_sum = DataSet(self)
        data_sum.data.extend(other.data)
        data_sum.ndata = self.ndata + other.ndata
        return data_sum

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        self.data.extend(other.data)
        self.ndata += other.ndata
        return self

    def append(self, profile):
        """Append data to an exisiting dataset."""
        if not isinstance(profile, base_profile.BaseProfile):
            raise TypeError
        self.data.append(profile)
        self.ndata += len(profile.val)

    def calculate_residuals(self, params, verbose=True, threshold=1e-3):
        """Calculate the residuals."""

        residuals = np.concatenate(
            [profile.calculate_residuals(params) for profile in self.data]
        )

        if verbose:
            chisq = sum(residuals ** 2)

            change = (chisq - self.chisq_ref) / self.chisq_ref

            if change < -threshold:
                nvarys = len([param for param in params.values() if param.vary])
                redchi = chisq / (self.ndata - nvarys)

                print("  * {:.3e} / {:.3e}".format(chisq, redchi))

                self.chisq_ref = chisq

        return residuals

    def write_to(self, params=None, output_dir="./"):
        """Write experimental and fitted profiles to a file."""
        datasets = dict()

        for profile in self.data:
            experiment_name = profile.experiment_name
            datasets.setdefault(experiment_name, list()).append(profile)

        for experiment_name, data in datasets.items():
            filename = "".join([experiment_name, ".dat"])
            filename = os.path.join(output_dir, filename)

            print("  * {}".format(filename))

            with open(filename, "w") as f:
                for profile in sorted(data, key=operator.attrgetter("peak")):
                    f.write(profile.print_profile(params=params))

    def add_dataset_from_file(self, filename, model=None, included=None, excluded=None):
        """Add data from a file to the dataset."""
        print("{:<45s} ".format(filename), end="")

        # Get the directory of the input file
        working_dir = os.path.dirname(filename)

        # Parse the experiment configuration file
        config = util.read_cfg_file(filename)

        try:
            # Read the experiment information
            details = dict(config.items("experiment"))
            experiment_type = details["type"]
            experiment_class = experiment_type.split(".")[0]

            # Read the experimental parameters
            details.update(
                {
                    key.lower(): val
                    for key, val in config.items("experimental_parameters")
                }
            )

            # Read the profile information (name, filename)
            filenames = {key.lower(): val for key, val in config.items("data")}

        except configparser.NoSectionError as e:
            sys.exit("    Reading aborted: {}".format(e))

        except KeyError as e:
            sys.exit(
                "\nIn the section 'experiment' of {}, '{}' must be provided!".format(
                    filename, e
                )
            )

        try:
            # Read (optional) additional parameters
            details.update(
                {key.lower(): val for key, val in config.items("extra_parameters")}
            )

        except configparser.NoSectionError:
            pass

        path = util.normalize_path(working_dir, details.get("path", "./"))

        try:
            reading = importlib.import_module(
                ".".join(["chemex.experiments", experiment_class, "reading"])
            )

        except ImportError:
            sys.exit(
                "The experiment '{}', referred in '{}' is not implemented.".format(
                    experiment_type, filename
                )
            )

        data, ndata = reading.read_profiles(
            path, filenames, details, model, included, excluded
        )

        self.data.extend(data)
        self.ndata += ndata

        print("{:<25s} {:<25d}".format(experiment_type, len(data)))

        return data



import itertools
import re

import lmfit as lf
from scipy import constants as cst

from chemex import parameters
from chemex.spindynamics import util


OTHER_SPIN = {"i": "s", "s": "i"}

RE_VARNAME = re.compile(
    "^(?P<name>.+?)(_(?P<spin>[is]{1,2}))?(_(?P<state>[abcd]{1,2}))?$"
)

DEFAULT_PARAMS = lf.Parameters()

for state1, state2 in itertools.combinations("abcd", 2):
    DEFAULT_PARAMS.add(f"k{state1}{state2}", value=0.0, min=0.0)

for state in "abcd":
    DEFAULT_PARAMS.add(f"p{state}", value=0.0, min=0.0, max=1.0)
    DEFAULT_PARAMS.add(f"j_{state}", value=0.0)
    DEFAULT_PARAMS.add(f"r1a_{state}", value=2.0, min=0.0)
    DEFAULT_PARAMS.add(f"sigma_{state}", value=0.0)
    DEFAULT_PARAMS.add(f"r2_mq_{state}", value=10.0, min=0.0)
    DEFAULT_PARAMS.add(f"mu_mq_{state}", value=0.0)

for spin, state in itertools.product("is", "abcd"):
    DEFAULT_PARAMS.add(f"cs_{spin}_{state}", value=0.0)
    DEFAULT_PARAMS.add(f"r1_{spin}_{state}", value=1.0, min=0.0)
    DEFAULT_PARAMS.add(f"r2_{spin}_{state}", value=10.0, min=0.0)
    DEFAULT_PARAMS.add(f"r2a_{spin}_{state}", value=12.0, min=0.0)
    DEFAULT_PARAMS.add(f"etaxy_{spin}_{state}", value=0.0)
    DEFAULT_PARAMS.add(f"etaz_{spin}_{state}", value=0.0)


def create_params(
    basis,
    model,
    nuclei=None,
    conditions=None,
    hn_ap_constraints=False,
    nh_constraints=False,
):
    """TODO"""

    fnames_k, params_k = create_params_k(model_name=model.name, conditions=conditions)
    fnames_lr, params_lr = create_params_lr(basis, conditions, nuclei)

    params_lr = set_thermal_factors(fnames_lr, params_lr, nuclei)

    if hn_ap_constraints:
        fnames_lr, params_lr = set_hn_ap_constraints(
            fnames_lr, params_lr, nuclei=nuclei, conditions=conditions
        )

    if nh_constraints:
        fnames_lr, params_lr = set_nh_constraints(fnames_lr, params_lr)

    fnames = dict(**fnames_k, **fnames_lr)

    # hack to add parameters depending on parameters not added yet
    params = lf.Parameters()
    params.add_many(*(lf.Parameter(fname) for fname in fnames.values()))
    params.update(params_lr)
    params.update(params_k)

    return fnames, params


def create_params_lr(basis, conditions, nuclei):
    names = get_default_names(basis)
    map_names = set_map_names(names, nuclei, conditions)
    params = buid_params(map_names)
    for name, full_name in map_names.items():
        if name.endswith(("_b", "_c", "_d")):
            name_a = "".join((name[:-1], "a"))
            full_name_a = map_names[name_a]
            params[full_name].set(expr=full_name_a)
    for state in ("b", "c", "d"):
        short_name = "cs_{}".format(state)
        short_name_expr = "cs_a + dw_a{}".format(state)
        parameters.set_param_expr(params, short_name, short_name_expr)
    return map_names, params


def get_default_names(basis):
    """TODO"""

    names = set()
    for name in basis._matrices:

        name_parsed = RE_VARNAME.match(name).groupdict()

        if not name.startswith(("k", "p", "w1", "carrier", "j_eff", "w_")):
            names.add(name)

            if name_parsed["name"] == "cs" and name_parsed["state"] != "a":
                name_dw = "dw_{spin}_a{state}".format(**name_parsed)
                names.add(name_dw)

    return names


def set_map_names(names=None, nuclei=None, conditions=None):
    """TODO"""

    if names is None:
        names = {}

    map_names = {}

    for name in names:
        name_parsed = RE_VARNAME.match(name).groupdict()

        short_name = "_".join(
            name_parsed[key]
            for key in ("name", "state")
            if name_parsed[key] is not None
        )

        definition = {"temperature": conditions.get("temperature", 0.0)}

        if name.startswith(("k", "p")):
            definition["p_total"] = conditions.get("p_total", 1.0)
            definition["l_total"] = conditions.get("l_total", 1.0)
        else:
            spin = name_parsed["spin"]
            if spin is None:
                spin = "is"

            definition["nuclei"] = nuclei[spin]

        if not name.startswith(("dw", "cs", "j")):
            definition["h_larmor_frq"] = conditions.get("h_larmor_frq", 1.0)

        map_names[name] = parameters.ParameterName(
            short_name, **definition
        ).to_full_name()

    return map_names


def buid_params(map_names):
    """TODO"""

    params = lf.Parameters()

    for name, full_name in map_names.items():
        param = lf.Parameter(full_name, value=0.0, vary=False)

        if name.startswith("p"):
            param.set(min=0.0, max=1.0)
        elif name.startswith("k"):
            param.set(min=0.0)
        elif name.startswith("r2"):
            param.set(value=10.0, min=0.0)
        elif name.startswith("r1"):
            param.set(value=1.0, min=0.0)
        elif name == "j":
            param.set(value=-93.0)

        params[full_name] = param

    return params


def set_thermal_factors(fnames, params, nuclei):
    """TODO"""

    for spin, state in itertools.product("is", "abcd"):

        theta = "theta_{}_{}".format(spin, state)

        if theta in fnames:

            r1 = "r1_{}_{}".format(spin, state)
            sigma = "sigma_{}".format(state)
            pop = "p{}".format(state)

            terms = []

            terms.append(
                "{g_ratio} * {r1}".format(
                    g_ratio=nuclei.atoms[spin]["g_ratio"], r1=fnames[r1]
                )
            )

            if sigma in fnames:
                terms.append(
                    "{g_ratio} * {sigma}".format(
                        g_ratio=nuclei.atoms[OTHER_SPIN[spin]]["g_ratio"],
                        sigma=fnames[sigma],
                    )
                )

            expr = "{pop} * ({sum})".format(pop=fnames[pop], sum=" + ".join(terms))

            params[fnames[theta]].set(expr=expr)

    for state in "abcd":

        theta = "theta_is_{}".format(state)

        if theta in fnames:

            pop = "p{}".format(state)

            terms = []

            for spin in "is":

                etaz = "etaz_{}_{}".format(spin, state)

                if etaz in fnames:
                    terms.append(
                        "{gamma} * {etaz}".format(
                            gamma=nuclei.atoms[spin]["g_ratio"], etaz=fnames[etaz]
                        )
                    )

            expr = "{pop} * ({sum})".format(pop=fnames[pop], sum=" + ".join(terms))

            params[fnames[theta]].set(expr=expr)

    return params


def set_hn_ap_constraints(map_names, params, nuclei=None, conditions=None):

    try:
        name_s = nuclei["s"]
    except IndexError:
        raise

    definition = {"nuclei": name_s}
    definition.update(conditions)

    for state in "abcd":

        r1_i = "r1_i_{}".format(state)
        r2_i = "r2_i_{}".format(state)
        r1_s = "r1_s_{}".format(state)
        r2a_i = "r2a_i_{}".format(state)
        r1a = "r1a_{}".format(state)
        r1 = "r1_{}".format(state)

        if r1_i in map_names:

            if r1_s not in map_names:

                map_names[r1_s] = parameters.ParameterName(
                    r1, **definition
                ).to_full_name()

                if state != "a":
                    expr = map_names["r1_s_a"]
                else:
                    expr = None

                params[map_names[r1_s]] = lf.Parameter(
                    map_names[r1_s], value=1.0, min=0.0, vary=False, expr=expr
                )

            params[map_names[r1_i]].set(
                expr="{} - {}".format(map_names[r1a], map_names[r1_s])
            )

            params[map_names[r2a_i]].set(
                expr="{} - {}".format(map_names[r2_i], map_names[r1_s])
            )

    return map_names, params


def set_nh_constraints(map_names, params):

    for state in "abcd":
        r1_i = "r1_i_{}".format(state)
        r2_i = "r2_i_{}".format(state)
        r2a_i = "r2a_i_{}".format(state)
        r1a = "r1a_{}".format(state)

        if r2a_i in map_names:
            params[map_names[r2a_i]].set(
                expr="{} + {} - {}".format(
                    map_names[r2_i], map_names[r1a], map_names[r1_i]
                )
            )

    return map_names, params


def build_global_name(shortname, attributes=None):
    if attributes is None:
        attributes = {}
    return parameters.ParameterName(shortname, **attributes).to_full_name()


def create_params_k(model_name=None, conditions=None):
    """Update the experimental and fitting parameters depending on the model.

    """

    fnames = {}
    params = lf.Parameters()

    available_model = (
        "2st.pb_kex",
        "2st.eyring",
        "2st.binding",
        "3st.pb_kex",
        "4st.pb_kex",
    )

    if model_name not in available_model:
        print("Warning: The 'model' option should either be:")
        for model_name in available_model:
            print("    - '{}'".format(model_name))
        print("Set to the default value: '2st.pb_kex'.")
        model_name = "2st.pb_kex"

    model = util.parse_model(model_name)
    attributes = {
        key: conditions.get(key, None) for key in {"temperature", "p_total", "l_total"}
    }

    states_all = "abcd"
    states = states_all[: model.state_nb]

    for pairs in itertools.permutations(states, 2):
        name = "k{}{}".format(*pairs)
        name_global = build_global_name(name, attributes)
        fnames[name] = name_global
        params.add(name=name_global, min=0.0, vary=False)

    for state in states:
        name = "p{}".format(state)
        name_global = build_global_name(name, attributes)
        fnames[name] = name_global
        params.add(name=name_global, min=0.0, max=1.0, vary=False)

    value_dict = {}
    vary_list = []
    extra = {}
    expr_dict = {}

    if model_name == "2st.pb_kex":

        fnames["kex_ab"] = build_global_name("kex_ab", attributes)
        params.add(name=fnames["kex_ab"], min=0.0, vary=False)

        value_dict = {"pb": 0.05, "kex_ab": 200.0}

        vary_list = ("pb", "kex_ab")

        expr_dict = {
            "pa": "1.0 - {pb}",
            "kab": "{kex_ab} * {pb}",
            "kba": "{kex_ab} * {pa}",
        }

    elif model_name == "2st.eyring":

        attributes = {key: conditions.get(key, None) for key in {"p_total", "l_total"}}

        for name in ("dh_b", "ds_b", "dh_ab", "ds_ab"):
            fname = build_global_name(name, attributes=attributes)
            fnames[name] = fname
            params.add(name=fname, vary=False)

        value_dict = {
            "dh_b": 6.5e+03,
            "dh_ab": 6.5e+04,
            "ds_b": 1.0e+10,
            "ds_ab": 1.0e+10,
        }

        vary_list = ("dh_b", "dh_ab")

        tk = conditions.get("temperature", 0.0) + 273.15
        kbt_h = cst.k * tk / cst.h
        rt = cst.R * tk

        extra = {"tk": tk, "kbt_h": kbt_h, "rt": rt}

        expr_dict = {
            "kab": "{kbt_h} * exp(-({dh_ab} - {tk} * {ds_ab}) / {rt})",
            "kba": "{kbt_h} * exp(-(({dh_ab} - {dh_b}) - {tk} * ({ds_ab} - {ds_b})) / {rt})",
            "pa": "{kba} / ({kba} + {kab})",
            "pb": "{kab} / ({kba} + {kab})",
        }

    elif model_name == "2st.binding":

        attributes = {"temperature": conditions.get("temperature", None)}

        for name in ("kd", "kon", "koff"):
            fname = build_global_name(name, attributes=attributes)
            fnames[name] = fname
            params.add(name=fname, min=0.0, vary=False)

        value_dict = {"kd": 1e-6, "kon": 1e7, "koff": 10.0}

        vary_list = ("kon", "koff")

        p_total = conditions.get("p_total", 0.0)
        l_total = conditions.get("l_total", 0.0)
        delta = l_total - p_total

        extra = {"delta": delta, "l_total": l_total}

        expr_dict = {
            "kd": "{koff} / {kon}",
            "kab": "{kon} * 0.5 * ({delta} - {kd} + sqrt(({delta} - {kd}) ** 2 + 4.0 * {kd} * {l_total}))",
            "kba": "{koff}",
            "pa": "{kba} / ({kba} + {kab})",
            "pb": "{kab} / ({kba} + {kab})",
        }

    elif model_name == "3st.pb_kex":

        for name in ("kex_ab", "kex_ac", "kex_bc"):
            fname = build_global_name(name, attributes)
            fnames[name] = fname
            params.add(name=fname, min=0.0)

        value_dict = {
            "pb": 0.02,
            "pc": 0.02,
            "kex_ab": 200.0,
            "kex_ac": 0.0,
            "kex_bc": 200.0,
        }

        vary_list = ("pb", "pc", "kex_ab", "kex_bc")

        expr_dict = {
            "pa": "1.0 - {pb} - {pc}",
            "kab": "{kex_ab} * {pb} / ({pa} + {pb}) if {pa} + {pb} else 0.0",
            "kba": "{kex_ab} * {pa} / ({pa} + {pb}) if {pa} + {pb} else 0.0",
            "kac": "{kex_ac} * {pc} / ({pa} + {pc}) if {pa} + {pc} else 0.0",
            "kca": "{kex_ac} * {pa} / ({pa} + {pc}) if {pa} + {pc} else 0.0",
            "kbc": "{kex_bc} * {pc} / ({pb} + {pc}) if {pb} + {pc} else 0.0",
            "kcb": "{kex_bc} * {pb} / ({pb} + {pc}) if {pb} + {pc} else 0.0",
        }

    elif model_name == "3st.eyring":

        attributes = {key: conditions.get(key, None) for key in {"p_total", "l_total"}}

        for name in (
            "dh_b",
            "ds_b",
            "dh_c",
            "ds_c",
            "dh_ab",
            "ds_ab",
            "dh_ac",
            "ds_ac",
            "dh_bc",
            "ds_bc",
        ):
            fname = build_global_name(name, attributes=attributes)
            fnames[name] = fname
            params.add(name=fname, vary=False)

        value_dict = {
            "dh_b": 6.5e+03,
            "dh_c": 6.5e+03,
            "dh_ab": 6.5e+04,
            "dh_ac": 1.0e+10,
            "dh_bc": 6.5e+04,
            "ds_b": 1.0e+10,
            "ds_c": 1.0e+10,
            "ds_ab": 1.0e+10,
            "ds_ac": 1.0e+10,
            "ds_bc": 1.0e+10,
        }

        vary_list = ("dh_b", "dh_c", "dh_ab", "dh_bc")

        tk = conditions.get("temperature", 0.0) + 273.15
        kbt_h = cst.k * tk / cst.h
        rt = cst.R * tk

        extra = {"tk": tk, "kbt_h": kbt_h, "rt": rt}

        expr_dict = {
            "kab": "{kbt_h} * exp(-({dh_ab} - {tk} * {ds_ab}) / {rt})",
            "kba": "{kbt_h} * exp(-(({dh_ab} - {dh_b}) - {tk} * ({ds_ab} - {ds_b})) / {rt})",
            "kac": "{kbt_h} * exp(-(({dh_ac} - {dh_a}) - {tk} * ({ds_ac} - {ds_a})) / {rt})",
            "kca": "{kbt_h} * exp(-(({dh_ac} - {dh_c}) - {tk} * ({ds_ac} - {ds_c})) / {rt})",
            "kbc": "{kbt_h} * exp(-(({dh_bc} - {dh_b}) - {tk} * ({ds_bc} - {ds_b})) / {rt})",
            "kcb": "{kbt_h} * exp(-(({dh_bc} - {dh_c}) - {tk} * ({ds_bc} - {ds_c})) / {rt})",
            "pa": "{kba} * {kca} / ({kba} * {kca} + {kab} * {kca} + {kac} * {kba})",
            "pb": "{kab} * {kcb} / ({kab} * {kcb} + {kba} * {kcb} + {kbc} * {kab})",
            "pc": "{kbc} * {kac} / ({kbc} * {kac} + {kcb} * {kab} + {kca} * {kbc})",
        }

    elif model_name == "4st.pb_kex":

        for name in ("kex_ab", "kex_ac", "kex_ad", "kex_bc", "kex_bd", "kex_cd"):
            fname = build_global_name(name, attributes)
            fnames[name] = fname
            params.add(name=fname, min=0.0)

        value_dict = {
            "pb": 0.02,
            "pc": 0.02,
            "pd": 0.02,
            "kex_ab": 200.0,
            "kex_ac": 0.0,
            "kex_ad": 0.0,
            "kex_bc": 200.0,
            "kex_bd": 0.0,
            "kex_cd": 200.0,
        }

        vary_list = ("pb", "pc", "pd", "kex_ab", "kex_bc", "kex_cd")

        expr_dict = {
            "pa": "1.0 - {pb} - {pc} - {pd}",
            "kab": "{kex_ab} * {pb} / ({pa} + {pb}) if {pa} + {pb} else 0.0",
            "kba": "{kex_ab} * {pa} / ({pa} + {pb}) if {pa} + {pb} else 0.0",
            "kac": "{kex_ac} * {pc} / ({pa} + {pc}) if {pa} + {pc} else 0.0",
            "kca": "{kex_ac} * {pa} / ({pa} + {pc}) if {pa} + {pc} else 0.0",
            "kad": "{kex_ad} * {pd} / ({pa} + {pd}) if {pa} + {pd} else 0.0",
            "kda": "{kex_ad} * {pa} / ({pa} + {pd}) if {pa} + {pd} else 0.0",
            "kbc": "{kex_bc} * {pc} / ({pb} + {pc}) if {pb} + {pc} else 0.0",
            "kcb": "{kex_bc} * {pb} / ({pb} + {pc}) if {pb} + {pc} else 0.0",
            "kbd": "{kex_bd} * {pd} / ({pb} + {pd}) if {pb} + {pd} else 0.0",
            "kdb": "{kex_bd} * {pb} / ({pb} + {pd}) if {pb} + {pd} else 0.0",
            "kcd": "{kex_cd} * {pd} / ({pc} + {pd}) if {pc} + {pd} else 0.0",
            "kdc": "{kex_cd} * {pc} / ({pc} + {pd}) if {pc} + {pd} else 0.0",
        }

    for name, value in value_dict.items():
        fname = fnames[name]
        params[fname].value = value

    for name in vary_list:
        fname = fnames[name]
        params[fname].vary = True

    for name, expr in expr_dict.items():
        fname = fnames[name]
        params[fname].expr = expr.format(**fnames, **extra)

    return fnames, params
